#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
import logging
logger = logging.getLogger('PyAMS (scheduler)')

from datetime import datetime

# import interfaces
from pyams_form.interfaces.form import IWidgetsSuffixViewletsManager
from pyams_scheduler.interfaces import IScheduler, SCHEDULER_HANDLER_KEY
from pyams_scheduler.zmi.interfaces import ISchedulerMenu
from pyams_skin.interfaces import IInnerPage, IPageHeader
from pyams_skin.interfaces.container import ITableElementEditor
from pyams_skin.layer import IPyAMSLayer
from pyams_utils.interfaces import VIEW_SYSTEM_PERMISSION, MANAGE_SYSTEM_PERMISSION
from pyams_zmi.interfaces.menu import IControlPanelMenu
from pyams_zmi.layer import IAdminLayer
from z3c.table.interfaces import IColumn, IValues
from zope.component.interfaces import ISite

# import packages
from pyams_form.form import AJAXEditForm
from pyams_pagelet.pagelet import pagelet_config
from pyams_skin.container import ContainerView
from pyams_skin.page import DefaultPageHeaderAdapter
from pyams_skin.table import DefaultElementEditorAdapter, BaseTable, I18nColumn, TrashColumn, ActionColumn
from pyams_skin.viewlet.menu import MenuItem
from pyams_template.template import template_config
from pyams_utils.adapter import adapter_config, ContextRequestViewAdapter
from pyams_utils.date import format_datetime
from pyams_utils.registry import query_utility
from pyams_utils.traversing import get_parent
from pyams_utils.url import absolute_url
from pyams_viewlet.manager import viewletmanager_config
from pyams_viewlet.viewlet import viewlet_config, Viewlet
from pyams_zmi.form import AdminDialogEditForm, AdminDialogDisplayForm
from pyams_zmi.view import AdminView
from pyams_zmq.socket import zmq_socket, zmq_response
from pyramid.url import resource_url
from pyramid.view import view_config
from z3c.form import field
from z3c.table.column import GetAttrColumn
from zope.interface import implementer, Interface

from pyams_scheduler import _


@adapter_config(context=(IScheduler, IAdminLayer, Interface), provides=ITableElementEditor)
class SchedulerTableElementEditor(DefaultElementEditorAdapter):
    """Scheduler table element editor"""

    view_name = 'scheduler-tasks.html'
    modal_target = False

    @property
    def url(self):
        site = get_parent(self.context, ISite)
        return resource_url(site, self.request, 'admin#{0}'.format(self.view_name))


@viewlet_config(name='scheduler.menu', context=ISite, layer=IAdminLayer, manager=IControlPanelMenu,
                permission=VIEW_SYSTEM_PERMISSION, weight=10)
@viewletmanager_config(name='scheduler.menu', context=ISite, layer=IAdminLayer)
@implementer(ISchedulerMenu)
class SchedulerMenuItem(MenuItem):
    """Scheduler menu"""

    label = _("Tasks scheduler")
    icon_class = 'fa-clock-o'
    url = '#scheduler-tasks.html'


class SchedulerTasksTable(BaseTable):
    """Scheduler tasks table"""

    id = 'scheduler_tasks_table'
    title = _("Scheduled tasks")
    cssClasses = {'table': 'table table-bordered table-striped table-hover table-tight datatable'}

    @property
    def data_attributes(self):
        manager = query_utility(IScheduler)
        attributes = super(SchedulerTasksTable, self).data_attributes
        table_attrs = {'data-ams-location': absolute_url(manager, self.request),
                       'data-ams-delete-target': 'delete-task.json'}
        if 'table' in attributes:
            attributes['table'].update(table_attrs)
        else:
            attributes['table'] = table_attrs
        return attributes

    def getCSSHighlightClass(self, column, item, cssClass):
        if column.__name__ == 'name':
            info = item.get_scheduling_info(self.request.registry)
            if info and not info.active:
                cssClass = 'text-dashed'
        return super(SchedulerTasksTable, self).getCSSHighlightClass(column, item, cssClass)


@adapter_config(name='settings', context=(Interface, IAdminLayer, SchedulerTasksTable), provides=IColumn)
class SchedulerTasksSettingsColumn(ActionColumn):
    """Scheduler tasks settings column"""

    icon_class = 'fa fa-fw fa-edit'
    icon_hint = _("Task settings")
    url = 'settings.html'
    target = None
    modal_target = True

    permission = VIEW_SYSTEM_PERMISSION
    weight = 1


@adapter_config(name='scheduler', context=(Interface, IAdminLayer, SchedulerTasksTable), provides=IColumn)
class SchedulerTasksScheduleColumn(ActionColumn):
    """Scheduler tasks schedule column"""

    icon_class = 'fa fa-fw fa-calendar'
    icon_hint = _("Schedule task")
    url = 'schedule.html'
    target = None
    modal_target = True

    permission = VIEW_SYSTEM_PERMISSION
    weight = 2


@adapter_config(name='name', context=(Interface, IAdminLayer, SchedulerTasksTable), provides=IColumn)
class SchedulerTasksNameColumn(I18nColumn, GetAttrColumn):
    """Scheduler tasks name column"""

    _header = _("Name")
    attrName = 'name'
    weight = 10


@adapter_config(name='run', context=(Interface, IAdminLayer, SchedulerTasksTable), provides=IColumn)
class SchedulerTasksRunColumn(ActionColumn):
    """Scheduler tasks run column"""

    icon_class = 'fa fa-fw fa-play'
    icon_hint = _("Run task")
    url = 'run.html'
    target = None
    modal_target = True

    permission = MANAGE_SYSTEM_PERMISSION
    weight = 20


@adapter_config(name='history', context=(Interface, IAdminLayer, SchedulerTasksTable), provides=IColumn)
class SchedulerTasksHistoryColumn(ActionColumn):
    """Scheduler tasks history column"""

    icon_class = 'fa fa-fw fa-history'
    icon_hint = _("Task run history")
    url = 'history.html'
    target = None
    modal_target = True

    permission = VIEW_SYSTEM_PERMISSION
    weight = 30


@adapter_config(name='trash', context=(Interface, IAdminLayer, SchedulerTasksTable), provides=IColumn)
class SchedulerTasksTrashColumn(TrashColumn):
    """Scheduler tasks trash column"""

    icon_hint = _("Delete task")
    permission = MANAGE_SYSTEM_PERMISSION


@view_config(name='delete-task.json', context=IScheduler, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
def delete_scheduler_task(request):
    """Delete task from scheduler utility"""
    translate = request.localizer.translate
    name = request.params.get('object_name')
    if not name:
        return {'status': 'message',
                'messagebox': {'status': 'error',
                               'content': translate(_("No provided object_name argument!"))}}
    if name not in request.context:
        return {'status': 'message',
                'messagebox': {'status': 'error',
                               'content': translate(_("Given task name doesn't exist!"))}}
    del request.context[name]
    return {'status': 'success'}


@adapter_config(context=(ISite, IAdminLayer, SchedulerTasksTable), provides=IValues)
class SchedulerTasksValuesAdapter(ContextRequestViewAdapter):
    """Scheduler tasks values adapter"""

    @property
    def values(self):
        manager = query_utility(IScheduler)
        if manager is not None:
            return list(manager.values())
        return ()


@pagelet_config(name='scheduler-tasks.html', context=ISite, layer=IPyAMSLayer, permission=VIEW_SYSTEM_PERMISSION)
@implementer(IInnerPage)
class SchedulerTasksView(AdminView, ContainerView):
    """Scheduler tasks view"""

    table_class = SchedulerTasksTable

    def __init__(self, context, request):
        super(SchedulerTasksView, self).__init__(context, request)


@adapter_config(context=(ISite, IAdminLayer, SchedulerTasksView), provides=IPageHeader)
class SchedulerTasksHeaderAdapter(DefaultPageHeaderAdapter):
    """Scheduler tasks header adapter"""

    icon_class = 'fa fa-fw fa-clock-o'
    title = _("Control panel")
    subtitle = _("Tasks scheduler")


@viewlet_config(name='scheduler.properties.menu', context=ISite, layer=IAdminLayer,
                manager=ISchedulerMenu, permission=VIEW_SYSTEM_PERMISSION, weight=1)
class SchedulerPropertiesMenuItem(MenuItem):
    """Scheduler properties menu"""

    label = _("Properties...")
    url = 'properties.html'
    modal_target = True

    def get_url(self):
        manager = query_utility(IScheduler)
        return resource_url(manager, self.request, self.url)


@pagelet_config(name='properties.html', context=IScheduler, layer=IPyAMSLayer, permission=VIEW_SYSTEM_PERMISSION)
class SchedulerPropertiesEditForm(AdminDialogEditForm):
    """Scheduler properties edit form"""

    @property
    def title(self):
        return self.context.__name__

    legend = _("Update tasks scheduler properties")

    fields = field.Fields(IScheduler).omit('__parent__', '__name__', 'tasks', 'history')
    ajax_handler = 'properties.json'
    edit_permission = MANAGE_SYSTEM_PERMISSION


@view_config(name='properties.json', context=IScheduler, request_type=IPyAMSLayer,
             permission=MANAGE_SYSTEM_PERMISSION, renderer='json', xhr=True)
class SchedulerPropertiesAJAXEditForm(AJAXEditForm, SchedulerPropertiesEditForm):
    """Scheduler properties edit form, AJAX view"""


@viewlet_config(name='scheduler.jobs.menu', context=ISite, layer=IAdminLayer,
                manager=ISchedulerMenu, permission=VIEW_SYSTEM_PERMISSION, weight=5)
class SchedulerJobsMenuItem(MenuItem):
    """Scheduler jobs menu"""

    label = _("Active jobs...")
    url = 'jobs.html'
    modal_target = True

    def get_url(self):
        manager = query_utility(IScheduler)
        return resource_url(manager, self.request, self.url)


@pagelet_config(name='jobs.html', context=IScheduler, layer=IPyAMSLayer, permission=VIEW_SYSTEM_PERMISSION)
class SchedulerJobsDisplayForm(AdminDialogDisplayForm):
    """Scheduler jobs display form"""

    @property
    def title(self):
        return self.context.__name__

    dialog_class = 'modal-large'
    legend = _("Display scheduler active jobs")

    fields = field.Fields(Interface)


class SchedulerJobsTable(BaseTable):
    """Scheduler jobs table"""

    title = _("Scheduler jobs")
    cssClasses = {'table': 'table table-bordered table-striped table-hover table-tight datatable'}
    sortOn = None

    @property
    def data_attributes(self):
        return {'table': {'data-ams-datatable-global-filter': 'false',
                          'data-ams-datatable-info': 'false',
                          'data-ams-datatable-sdom': "t<'dt-row dt-bottom-row'<'text-right'p>>",
                          'data-ams-datatable-display-length': '20',
                          'data-ams-datatable-pagination-type': 'bootstrap_prevnext'}}


@adapter_config(name='name', context=(Interface, IAdminLayer, SchedulerJobsTable), provides=IColumn)
class SchedulerJobNameColumn(I18nColumn, GetAttrColumn):
    """Scheduler job name column"""

    _header = _("Job name")
    weight = 1

    def getValue(self, obj):
        return obj['name']


@adapter_config(name='id', context=(Interface, IAdminLayer, SchedulerJobsTable), provides=IColumn)
class SchedulerJobIdColumn(I18nColumn, GetAttrColumn):
    """Scheduler job ID column"""

    _header = _("ID")
    weight = 10

    def getValue(self, obj):
        return obj['id']


@adapter_config(name='trigger', context=(Interface, IAdminLayer, SchedulerJobsTable), provides=IColumn)
class SchedulerJobTriggerColumn(I18nColumn, GetAttrColumn):
    """Scheduler job trigger column"""

    _header = _("Trigger")
    weight = 20

    def getValue(self, obj):
        return obj['trigger']


@adapter_config(name='next_run', context=(Interface, IAdminLayer, SchedulerJobsTable), provides=IColumn)
class SchedulerJobNextRunColumn(I18nColumn, GetAttrColumn):
    """Scheduler job next run column"""

    _header = _("Next run")
    weight = 30

    def getValue(self, obj):
        return format_datetime(datetime.utcfromtimestamp(obj['next_run']), request=self.request)


@adapter_config(context=(IScheduler, IAdminLayer, SchedulerJobsTable), provides=IValues)
class SchedulerJobsValuesAdapter(ContextRequestViewAdapter):
    """Scheduler jobs values adapter"""

    @property
    def values(self):
        handler = self.request.registry.settings.get(SCHEDULER_HANDLER_KEY, False)
        if handler:
            socket = zmq_socket(handler)
            socket.send_json(['get_jobs', {}])
            status, response = zmq_response(socket)
            if status == 200:
                return response
            else:  # error
                return ()
        else:
            return ()


@viewlet_config(name='scheduler-jobs', view=SchedulerJobsDisplayForm, layer=IAdminLayer,
                manager=IWidgetsSuffixViewletsManager)
@template_config(template='templates/scheduler-jobs.pt')
class SchedulerJobsViewlet(Viewlet):
    """Scheduler jobs viewlet"""

    table = SchedulerJobsTable

    def __init__(self, context, request, view, manager):
        super(SchedulerJobsViewlet, self).__init__(context, request, view, manager)
        self.table = SchedulerJobsTable(context, request)

    def update(self):
        super(SchedulerJobsViewlet, self).update()
        self.table.update()
